""" 
Helper classes
"""


import sys


class UsePath():
    """Quick class to use as a context manager when importing from a relative path. While in this manager's context,
    avoid making changes to sys.path, as this manager inserts a value at the first position on enter, and removes
    whichever element is in that position on exit.
    """

    def __init__(self, path):
        """ Saves ``path`` to the UsePath instance """
        self.path = path

    def __enter__(self):
        """ Appends ``path`` to ``sys.path``, at index 0 (the start) """
        sys.path.insert(0, self.path)

    def __exit__(self, *args):
        """ Removes the item at ``sys.path``'s index 0 (should be ``path``). """
        sys.path.pop(0)


class Sensitive():
    """ Flags a setting as sensitive, indicating to the Settings object that
    this setting should be hidden when ``Settings.masked`` is invoked. """

    def __init__(self, obj, stars:int=10):
        """ Create a new ``Sensitive`` instance around ``obj``. You may optionally
        pass stars=(int > 0) the number of asterisks to display when
        Settings.masked is invoked (default 10). """

        self.obj = obj

        if stars <= 0:
            raise AttributeError('Stars must be greater than 0.')

        self.stars = stars

    def __str__(self):
        """ Returns the str() of the wrapped object """
        return str(self.obj)

    def __repr__(self):
        """ Returns "Sensitive: " and the repr() of the wrapped object """
        return f"Sensitive ({repr(self.obj)})"

    def mask(self):
        """ Returns the masked value of the object (a str of asterisks with 
        length equal to ``self.stars``)"""
        return "*" * self.stars



"""
Loader class
"""

import os
import re
from glob import glob
from importlib import import_module, reload

from enpyronments.settings import Settings
from enpyronments.utils import UsePath

default_prefix = "env"
default_ext = ".py"
default_sep = "_"
default_builtin_pattern = r"^__(.+)__$"
default_attribute_pattern = r"^[A-Z_0-9]+$"
default_local_name = "local"
default_mode_name = "MODE"


class Loader:
    """The settings loader.follows a set resolution order to determin which environment settings will be loaded

    Keyword Arguments:
        root {str} -- Path representing the root directory to start importing from

        prefix {str} -- Prefix environment settings files will have

        ext {str} -- Extension environment settings files will have

        sep {str} -- Separator between environment specifiers

        builtin_pattern {str} -- Regex pattern used to identify which attributes are builtins, which should be ignored

        attribute_pattern {str} -- Regex pattern used to identify which attributes should be included

        mode {str} -- Name of setting indicating mode

        local_name -- Suffix used to indicate local settings
    """

    def __init__(
        self,
        root,
        prefix=default_prefix,
        ext=default_ext,
        sep=default_sep,
        builtin_pattern=default_builtin_pattern,
        attribute_pattern=default_attribute_pattern,
        mode=default_mode_name,
        local_name=default_local_name,
    ):
        self.root = root
        self.prefix = prefix
        self.ext = ext
        self.sep = sep
        self.builtin_pattern = builtin_pattern
        self.attribute_pattern = attribute_pattern
        self.mode = mode
        self.local_name = local_name

    def find_modules(self, package):
        """Searches for modules in package that match the glob pattern 
        {prefix}*{ext}
        
        Arguments:
            package {str} -- package from which to import modules
        """
        pattern = os.path.join(self.root, package, f"{self.prefix}*{self.ext}")
        with UsePath(self.root):
            for module_path in glob(pattern):
                module_name = module_path.split(os.path.sep)[-1].replace(self.ext, "")
                module = import_module(f"{package}.{module_name}")
                # Reload module, to ensure same-named packages don't interfere
                self.refresh_module(module)
                name = module.__name__.split(".")[-1]
                module_dict = dict(self.get_module_attrs(module))
                yield name, module_dict

    def refresh_module(self, module):
        """ Imports current settings from a module name... by force """
        keys = list(module.__dict__.keys())
        for key in keys:
            if not re.match(self.builtin_pattern, key):
                delattr(module, key)
        reload(module)
        return module

    def get_module_attrs(self, module):
        """Given a module object, extract the names of all attributes that match attribute_pattern,
        excluding those that match bulitin_pattern.
        
        Arguments:
            module {module} -- The module to extract attributes of
        """
        for key, val in module.__dict__.copy().items():
            # skip over attributes that look like builtins
            if re.match(self.builtin_pattern, key):
                continue
            # skip any attributes that don't look like our regex
            if not re.match(self.attribute_pattern, key):
                continue
            yield key, val

    def get_mode(self, settings_by_module):
        """Given a dictionary of settings, determine what mode to extract settings from
        
        Arguments:
            settings_by_module {dict} -- settings dictionary
        """

        def get_mode_setting(name):
            """ Helper function to pull the mode setting from the settings dict """
            if name in settings_by_module:
                return settings_by_module[name].get(self.mode)
            return None

        local_settings_module = f"{self.prefix}{self.sep}{self.local_name}"
        local_mode = get_mode_setting(local_settings_module)

        if local_mode:
            return local_mode

        global_settings_module = f"{self.prefix}"
        global_mode = get_mode_setting(global_settings_module)

        return global_mode

    def get_mode_settings(self, mode, settings_by_module):
        """Exclude settings for modes we're not using
        
        Arguments:
            mode {str} -- name of mode to load modules for
            settings_by_module {dict} -- settings dictionary
        """
        if not mode:
            return settings_by_module

        mode_settings = {}
        for key, val in settings_by_module.items():
            if (
                mode in key.split(self.sep)
                or key == self.prefix
                or key == self.sep.join([self.prefix, self.local_name])
            ):
                mode_settings[key] = val
        return mode_settings

    def get_load_order(self, mode):
        """ Defines the order in which settings are loaded. If needed, you can override this method in a subclass to
        define custom logic for your use case 
        
        Arguments:
            mode {str} -- name of mode to load
        """
        load_order = [
            self.sep.join([self.prefix]),
            self.sep.join([self.prefix, self.local_name]),
        ]
        if mode:
            load_order.append(self.sep.join([self.prefix, mode]))
            load_order.append(self.sep.join([self.prefix, mode, self.local_name]))

        return load_order

    def load_settings(self, package):
        """Load the settings files found in package, prioritizing local settings, then mode specific settings, then
        general settings (i.e. env_dev_local beats env_dev beats env_local beats env)
        
        Arguments:
            package {str} -- package name
        """

        settings_by_module = {}
        for module_name, module in self.find_modules(package):
            settings_by_module[module_name] = module

        mode = self.get_mode(settings_by_module)
        mode_settings = self.get_mode_settings(mode, settings_by_module)

        load_order = self.get_load_order(mode)

        settings = Settings()
        for key in load_order:
            new_settings = mode_settings.get(key)
            if new_settings:
                settings.update(new_settings)

        return settings



"""
Settings class
"""

import os
from collections.abc import MutableMapping

from enpyronments.utils import Sensitive


class Settings(MutableMapping):
    """Holder object for settings. Ensures that Sensitive values are masked when
    masked() is invoked, and otherwise return their values"""

    def __init__(self, iterable=None, **kwargs):
        """ Emulate dict.__init__'s style of accepting either an iterable of pairs or a set of keyword arguments """
        if kwargs:
            self.data = dict(**kwargs)
        elif iterable:
            self.data = dict(iterable)
        else:
            self.data = dict()

    def masked(self):
        return dict(
            zip(
                self.data.keys(),
                (
                    v.mask() if isinstance(v, Sensitive) else v
                    for v in self.values(extract_from_sensitive=False)
                ),
            )
        )

    def extract_from_sensitive(self, val, extract=True):
        if extract and isinstance(val, Sensitive):
            return val.obj
        return val

    def __getitem__(self, key, extract_from_sensitive: bool = True):
        """ same as dict.__getitem__, but extracts the value of Sensitive type elements """
        val = self.data.__getitem__(key)
        return self.extract_from_sensitive(val, extract_from_sensitive)

    def __getattr__(self, key):
        """ Attempts to return attributes set on self first, then on self.data """
        # Also lookup lowercase versions of settings
        if key not in self.data:
            key = str(key).upper()
        try:
            return self.extract_from_sensitive(self.data[key])
        except KeyError:
            raise KeyError(
                f'"{key}" was not found in specified settings and is not an attribute of {repr(self)}.'
            )

    def __setitem__(self, key, val):
        """ Same as dict.__setitem__, but sets the value of Sensitive type elements if the current value is already
        Sensitive """
        # Don't want to set a Sensitive value to a nonsensitive value, so if:
        # 1- The key we're setting is already set
        # 2- The value we already have is sensitive
        # 3- The value we're setting isn't already sensitive
        # Then we need to wrap val in Sensitive().

        if (
            key in self.data
            and isinstance(self.get(key, extract_from_sensitive=False), Sensitive)
            and not isinstance(val, Sensitive)
        ):
            return self.data.__setitem__(key, Sensitive(val))
        return self.data.__setitem__(key, val)

    def __delitem__(self, key):
        """ Same as dict.__delitem__ """
        return self.data.__delitem__(key)

    def __iter__(self):
        """ Same as dict.__iter__ """
        return self.data.__iter__()

    def __len__(self):
        """ Same as dict.__len__ """
        return self.data.__len__()

    def get(self, key, default=None, extract_from_sensitive: bool = True):
        """ Same as dict.get, but extracts the value of Sensitive type elements """
        val = self.data.get(key)
        return self.extract_from_sensitive(val, extract_from_sensitive)

    def items(self, extract_from_sensitive: bool = True) -> tuple:
        """ Same as dict.items, but extracts the value of Sensitive type elements """
        for key, val in self.data.items():
            if extract_from_sensitive and isinstance(val, Sensitive):
                yield key, val.obj
            yield key, val

    def values(self, extract_from_sensitive: bool = True) -> tuple:
        """ Same as dict.values, but extracts the value of Sensitive type elements """
        for val in self.data.values():
            if extract_from_sensitive and isinstance(val, Sensitive):
                yield val.obj
            yield val

    def keys(self):
        """ Returns the keys in self.data """
        return self.data.keys()

    def save_to_environ(self):
        """Saves the current state of settings to the environment via os.environ"""
        os.environ.update(self.data)
